@@ -15,7 +15,29 @@ use Time::Piece ();
 has 'mysql';
 has 'no_txn' => sub { 0 };
 
-our $VERSION = '0.31';
+our $VERSION = '0.32';
+
+# The dequeue system has a couple limitations:
+# 1. There is no way to directly notify a sleeping worker of an incoming
+#    job
+# 2. There is a race condition between identifying a runnable job
+#    and claiming it for the current worker.
+#
+# The first is solved by Mojo::mysql::PubSub, which currently makes
+# a new connection to MySQL, records the connection ID, and then sleeps. 
+# When a message is "published", the publisher writes the message to
+# a table and then `KILL`s all the sleeping "subscribers". One of the
+# subscribers reads the message from the table and continues.
+#
+# The second is solved by checking how many rows are `UPDATE`d when
+# claiming the job. If none, we try again to dequeue immediately. This
+# happens $DEQUEUE_RACE_ATTEMPTS times before giving up.
+#
+# When it gives up, there may still be jobs available to be dequeued. In
+# most cases, this is okay: The worker will start sleeping for an
+# incoming job and when one comes in, it will wake up and start
+# dequeuing again (even if the new job itself isn't ready to run).
+my $DEQUEUE_RACE_ATTEMPTS = 10;
 
 sub dequeue {
   my ($self, $worker_id, $wait, $options) = @_;
@@ -629,7 +651,7 @@ sub _try {
 
   warn "Dequeuing SQL: $sql" if DEBUG;
   my $db = $self->mysql->db;
-  my $i = 5; # Try this many times to get a job
+  my $i = $DEQUEUE_RACE_ATTEMPTS;
   my $job;
   while ( $i-- > 0 ) {
     # Find a candidate job to run
@@ -652,11 +674,8 @@ sub _try {
     last;
   }
 
-  #; use Data::Dumper;
-  #; say "Dequeued job: " . Dumper $job;
-
+  return undef unless $job;
   $job->{args} = $job->{args} ? decode_json($job->{args}) : undef;
-
   return $job;
 }
 
@@ -1175,7 +1194,7 @@ Minion::Backend::mysql
 
 =head1 VERSION
 
-version 0.31
+version 0.32
 
 =head1 SYNOPSIS
 
@@ -1664,27 +1683,27 @@ __DATA__
 @@ minion
 -- 1 up
 create table if not exists minion_jobs (
-		`id`       serial not null primary key,
-		`args`     mediumblob not null,
-		`created`  timestamp not null default current_timestamp,
-		`delayed`  timestamp not null default current_timestamp,
-		`finished` timestamp null,
-		`priority` int not null,
-		`result`   mediumblob,
-		`retried`  timestamp null,
-		`retries`  int not null default 0,
-		`started`  timestamp null,
-		`state`    varchar(128) not null default 'inactive',
-		`task`     text not null,
-		`worker`   bigint
+  `id`       serial not null primary key,
+  `args`     mediumblob not null,
+  `created`  timestamp not null default current_timestamp,
+  `delayed`  timestamp not null default current_timestamp,
+  `finished` timestamp null,
+  `priority` int not null,
+  `result`   mediumblob,
+  `retried`  timestamp null,
+  `retries`  int not null default 0,
+  `started`  timestamp null,
+  `state`    varchar(128) not null default 'inactive',
+  `task`     text not null,
+  `worker`   bigint
 );
 
 create table if not exists minion_workers (
-		`id`      serial not null primary key,
-		`host`    text not null,
-		`pid`     int not null,
-		`started` timestamp not null default current_timestamp,
-		`notified` timestamp not null default current_timestamp
+  `id`      serial not null primary key,
+  `host`    text not null,
+  `pid`     int not null,
+  `started` timestamp not null default current_timestamp,
+  `notified` timestamp not null default current_timestamp
 );
 
 -- 1 down
@@ -1806,11 +1825,11 @@ ALTER TABLE minion_jobs DROP COLUMN lax;
 
 -- 8 up
 ALTER TABLE minion_jobs_depends
-    ADD PRIMARY KEY (parent_id, child_id);
+  ADD PRIMARY KEY (parent_id, child_id);
 
 -- 8 down
 ALTER TABLE minion_jobs_depends
-    DROP PRIMARY KEY;
+  DROP PRIMARY KEY;
 
 -- 9 up
 DROP INDEX minion_jobs_state_idx ON minion_jobs;
@@ -1838,7 +1857,19 @@ ALTER TABLE minion_jobs_depends ADD COLUMN state ENUM('inactive','active','finis
 ALTER TABLE minion_jobs_depends ADD COLUMN expires DATETIME;
 UPDATE minion_jobs_depends dep JOIN minion_jobs job ON dep.parent_id=job.id SET dep.state=job.state, dep.expires=job.expires;
 CREATE TRIGGER minion_trigger_insert_depends BEFORE INSERT ON minion_jobs_depends FOR EACH ROW
-    SET NEW.state = COALESCE(( SELECT state FROM minion_jobs WHERE id=NEW.parent_id ), 'finished'),
-        NEW.expires = ( SELECT expires FROM minion_jobs WHERE id=NEW.parent_id );
+  SET NEW.state = COALESCE(( SELECT state FROM minion_jobs WHERE id=NEW.parent_id ), 'finished'),
+    NEW.expires = ( SELECT expires FROM minion_jobs WHERE id=NEW.parent_id );
 CREATE TRIGGER minion_trigger_update_jobs AFTER UPDATE ON minion_jobs FOR EACH ROW
-    UPDATE minion_jobs_depends SET state=NEW.state, expires=NEW.expires WHERE parent_id=OLD.id;
+  UPDATE minion_jobs_depends SET state=NEW.state, expires=NEW.expires WHERE parent_id=OLD.id;
+
+-- 13 up
+CREATE INDEX minion_jobs_depends_state_expires ON minion_jobs_depends (state, expires);
+
+# Found useless indexes with: SELECT * FROM sys.schema_unused_indexes
+DROP INDEX `minion_jobs_expires` ON `minion_jobs`;
+DROP INDEX `minion_notes_note_key` ON `minion_notes`;
+
+-- 13 down
+DROP INDEX minion_jobs_depends_state_expires ON minion_jobs_depends;
+CREATE INDEX minion_notes_note_key ON minion_notes (note_key);
+CREATE INDEX minion_jobs_expires ON minion_jobs (expires);
